var CAPTURE_DELAY = 150;

function onMessage(data, sender, callback) {
    if (data.msg === 'scrollPage') {
        getPositions(callback);
        return true;
    }
    return true;
}

if (!window.hasScreenCapturePage) {
    window.hasScreenCapturePage = true;
    chrome.runtime.onMessage.addListener(onMessage);
}

function max(nums) {
    return Math.max.apply(Math, nums.filter(function(x) { return x; }));
}

function getPositions(callback) {

    var floatingElements = getFloatingElements() || [];
    var rect;
    var rectElement = document.querySelector("#eagle-resize-tool .resizers");
    if (rectElement) {
        rect = {
            x: rectElement.offsetLeft,
            y: rectElement.offsetTop,
            width: rectElement.clientWidth,
            height: rectElement.clientHeight,
        };
    }
    else {
        floatingElements.forEach(function (elem) {
            $(elem).attr("eagle-scrolling-hidden", "true");
        });
    }

    var scrollElement = getMainScrollElement();
    if (rect) {
        scrollElement = document.documentElement;
    }

    var body = document.body;
    var originalY;
    if (!scrollElement || scrollElement === document.documentElement) {
        originalY = window.scrollY;
    }
    else {
        scrollElement.scrollTop;
    }

    var widths = [
            document.documentElement.clientWidth,
            body ? body.scrollWidth : 0,
            document.documentElement.scrollWidth,
            body ? body.offsetWidth : 0,
            document.documentElement.offsetWidth
        ],
        heights = [
            scrollElement && scrollElement.clientHeight || 0,
            scrollElement && scrollElement.scrollHeight || 0
        ],
        fullWidth = max(widths),
        fullHeight = max(heights),
        windowWidth = window.innerWidth,
        windowHeight = window.innerHeight,
        scrollPositions = [],
        scrollPad = 200,
        yDelta = windowHeight - (windowHeight > scrollPad ? scrollPad : 0),
        xDelta = windowWidth,
        // yPos = fullHeight - windowHeight,
        yPos = fullHeight,
        xPos,
        numOfScrollPostions;

    if (!scrollElement || scrollElement === document.documentElement) {
        heights = [
            document.documentElement.clientHeight,
            body ? body.scrollHeight : 0,
            document.documentElement.scrollHeight,
            body ? body.offsetHeight : 0,
            document.documentElement.offsetHeight,
        ]
    }
    else {
        heights = [
            scrollElement && scrollElement.clientHeight || 0,
            scrollElement && scrollElement.scrollHeight || 0
        ]
    }

    // During zooming, there can be weird off-by-1 types of things...
    if (fullWidth <= xDelta + 1) {
        fullWidth = xDelta;
    }

    if (fullWidth > window.innerWidth) {
        fullWidth = window.innerWidth;
    }

    // Disable all scrollbars. We'll restore the scrollbar state when we're done
    if (!rect) {
        document.documentElement.classList.add("eagle-scrolling");
    }

    document.documentElement.classList.add("hide-eagle-cropper");

    // 全屏幕截图
    if (!rect) {
        while (yPos > -yDelta) {
            xPos = 0;
            while (xPos < fullWidth) {
                scrollPositions.push([xPos, yPos]);
                xPos += xDelta;
            }
            yPos -= yDelta;
        }
    }
    // 手动选择截图区
    else {
        yPos = rect.y + rect.height;
        while (yPos > rect.y - yDelta) {
            scrollPositions.push([0, yPos]);
            yPos -= yDelta;
        }   
    }

    numOfScrollPostions = scrollPositions.length;

    var $metaTitle = jQuery('head title', window.parent.document);
    var originTitle = $metaTitle.text();

    function cleanUp() {
        $metaTitle.text(originTitle);
        document.documentElement.classList.remove("eagle-scrolling");
        document.documentElement.classList.remove("eagle-capturing");
        floatingElements.forEach(function (elem) {
            $(elem).removeAttr("eagle-scrolling-hidden");
        });
        if (!scrollElement || scrollElement === document.documentElement) {
            window.scrollTo(0, originalY);
        }
        else {
            scrollElement.scrollTop = originalY;
        }
    }

    var lastTime = Date.now();
    // 开始滚动，直到结束
    (function processArrangements() {

        if (!rect) {
            var floatingElements = getFloatingElements() || [];
            floatingElements.forEach(function (elem) {
                $(elem).attr("eagle-scrolling-hidden", "true");
            });
        }

        if (!scrollPositions.length) {
            cleanUp();
            if (callback) { callback(); }
            return;
        }

        var next = scrollPositions.shift(),
            x = next[0],
            y = next[1];

        document.documentElement.classList.remove("eagle-capturing");
        if (!rect) {
            // 如果 body html 有 overflow-x: hidden 会造成无法滚动
            if (!scrollElement || scrollElement === document.documentElement) {
                window.scrollTo(0, y);
            }
            else {
                scrollElement.scrollTop = y;
            }
        }
        else {
            if (!scrollElement || scrollElement === document.documentElement) {
                // 如果已经涵盖，就不需要滚动
                var scrollTop = jQuery(window).scrollTop();
                if (scrollTop < rect.y && scrollTop + windowHeight > rect.y + rect.height && rect.height < windowHeight) {

                }
                else {
                    window.scrollTo(0, y);
                }
            }
            else {
                // scrollElement.scrollTop = y;
            }
        }

        var complete = (numOfScrollPostions - scrollPositions.length) / numOfScrollPostions;
        var data = {
            msg: 'capture',
            x: window.scrollX,
            complete: complete,
            windowWidth: windowWidth,
            totalWidth: fullWidth,
            totalHeight: fullHeight,
            rect: rect,
            numOfScrollPostions: (numOfScrollPostions - scrollPositions.length),
            pixelRatio: window.devicePixelRatio
        };

        $metaTitle.text(`${parseInt(complete * 100)}% | ${originTitle}`);

        if (!scrollElement || scrollElement === document.documentElement) {
            data.y = window.scrollY;
        }
        else {
            data.y = scrollElement.scrollTop;
        }

        var delay = CAPTURE_DELAY;

        window.setTimeout(function() {
            var cleanUpTimeout = window.setTimeout(cleanUp, 2000);
            document.documentElement.classList.add("eagle-capturing");
            lastTime = Date.now();
            // processArrangements();
            window.clearTimeout(cleanUpTimeout);
            chrome.runtime.sendMessage(data, function(captured) {
                if (captured) {
                    if (scrollElement && scrollPositions) {
                        if (scrollElement.scrollTop !== data.y) {
                            setTimeout(function () {
                                processArrangements();
                            }, 1000);
                        }
                        else {
                            processArrangements();
                        }
                    }
                    else {
                        processArrangements();
                    }
                }
            });
        }, delay);
    })();
}

function getZIndex (e) {
    if (e==document) return 0;
    var z = document.defaultView.getComputedStyle(e).getPropertyValue('z-index');
    if (isNaN(z)) return getZIndex(e.parentNode);
    else return z;
};

function isElementVisible(elem) {
    try {
        if (elem.nodeName === "HTML" || elem.nodeName === "BODY") return true;
        const style = getComputedStyle(elem);
        if (style.display === 'none') return false;
        if (style.visibility !== 'visible') return false;
        if (style.opacity === 0) return false;
        if (elem.offsetWidth + elem.offsetHeight + elem.getBoundingClientRect().height +
            elem.getBoundingClientRect().width === 0) {
            return false;
        }
        var elementPoints = {
            'center': {
                x: elem.getBoundingClientRect().left + elem.offsetWidth / 2,
                y: elem.getBoundingClientRect().top + elem.offsetHeight / 2
            },
            'top-left': {
                x: elem.getBoundingClientRect().left,
                y: elem.getBoundingClientRect().top
            },
            'top-right': {
                x: elem.getBoundingClientRect().right,
                y: elem.getBoundingClientRect().top
            },
            'bottom-left': {
                x: elem.getBoundingClientRect().left,
                y: elem.getBoundingClientRect().bottom
            },
            'bottom-right': {
                x: elem.getBoundingClientRect().right,
                y: elem.getBoundingClientRect().bottom
            }
        }

        for(index in elementPoints) {
            var point = elementPoints[index];
            if (point.x < 0) return false;
            if (point.x > (document.documentElement.clientWidth || window.innerWidth)) return false;
            if (point.y < 0) return false;
            if (point.y > (document.documentElement.clientHeight || window.innerHeight)) return false;
            let pointContainer = document.elementFromPoint(point.x, point.y);
            if (pointContainer !== null) {
                do {
                    if (pointContainer === elem) return true;
                } while (pointContainer = pointContainer.parentNode);
            }
        }
        return false;
    }
    catch (err) {
        return true;
    }
}

function getMainScrollElement () {
    var scrollableElements = [];
    var all = document.getElementsByTagName("*");
    for (var i = 0; i < all.length; i++) {
        var elem = all[i];
        if (elem.scrollHeight > elem.clientHeight) {
            scrollableElements.push({
                elem: elem,
                scrollHeight: elem.scrollHeight,
                clientHeight: elem.clientHeight
            });
        }
    }

    scrollableElements = scrollableElements.filter(function (item) {
        if (item.clientHeight > window.innerHeight) { return false; }
        if (item.elem && item.elem.nodeName === "TEXTAREA") return false;
        if (!isElementVisible(item.elem)) return false;
        var overflow = getComputedStyle(item.elem)["overflow"];
        return overflow.includes("auto") || overflow.includes("scroll") ;
    })

    if (getComputedStyle(document.documentElement)["overflowY"] !== "hidden") {
        scrollableElements.unshift({
            scrollHeight: document.documentElement.scrollHeight,
            clientHeight: document.documentElement.clientHeight,
            elem: document.documentElement
        });
    }

    scrollableElements = scrollableElements.sort(function (a, b) {
        // 侦测 overlay scrollable div
        if (a.clientHeight === b.clientHeight && a.scrollHeight !== b.scrollHeight) {
            var aZindex = parseInt(getZIndex(a.elem));
            var bZindex = parseInt(getZIndex(b.elem));
            if (aZindex < bZindex)
                return 1;
            if (aZindex > bZindex)
                return -1;
            if (aZindex === 0 && bZindex === 0) {
                if (a.scrollHeight < b.scrollHeight)
                    return 1;
                else
                    return -1;
            }
            else {
                return 0;
            }
        }
        else {
            if (a.scrollHeight < b.scrollHeight)
                return 1;
            if (a.scrollHeight > b.scrollHeight)
                return -1;
            return 0;
        }
    });

    if (scrollableElements.length > 0) {
        return scrollableElements[0].elem;
    }
    return document.documentElement;
}

function getFloatingElements () {

    var all = document.getElementsByTagName("*");

    // 获取所有浮动元件
    var floatingElements = [];
    for (var i = 0; i < all.length; i++) {
        var elem = all[i];
        var style = getComputedStyle(elem);
        var position = style["position"];
        var isVisible = style.width !== "0" &&
                        style.height !== "0" &&
                        style.opacity !== "0" &&
                        style.display!=='none' &&
                        style.visibility!== 'hidden';
        if ( isVisible && (position.includes("fixed") || position.includes("sticky")) ) {
            floatingElements.push(elem)
        }
    }

    // 取出可疑者
    var annoyingElements = [];
    floatingElements.forEach(function (elem) {
        var $elem = jQuery(elem);
        var windowArea = window.innerHeight * window.innerWidth;
        if (isElementInViewport(elem) && $elem.is(":visible")) {
            var style = getComputedStyle(elem);
            if (
                (
                    $elem.height() * $elem.width() < windowArea * 0.2 && 
                    ($elem.position().top > 10 && $elem.width() < window.innerWidth * 0.8 )
                )
                ||
                (
                    style.bottom === "0px" && style.top !== "0px"
                )
                ||
                (
                    $elem.height() < 100 && $elem.width() < 100 && $elem.height() * $elem.width() < 5625
                )
            ) {
                annoyingElements.push(elem);
            }
        }
    });

    function isElementInViewport(el) {
        if (!el) return false;
        //special bonus for those using jQuery
        if (typeof jQuery === "function" && el instanceof jQuery) {
            el = el[0];
        }
        if (!el) {
            return false;
        }
        var rect = el.getBoundingClientRect();
        return (
            rect.top + rect.height >= 0 &&
            rect.left >= 0 &&
            rect.bottom - rect.height <= (window.innerHeight || document.documentElement.clientHeight) && /*or $(window).height() */
            rect.right <= (window.innerWidth || document.documentElement.clientWidth) /*or $(window).width() */
        );
    }

    return annoyingElements;

}